import sbol2
SBH_USER = 'sd2e'
SBH_PASSWORD = 'jWJ1yztJl2f7RaePHMtXmxBBHwNt'

sbh = sbol2.PartShop('https://hub.sd2e.org')
sbh.login(SBH_USER, SBH_PASSWORD)
doc = sbol2.Document()
uri = ('https://hub.sd2e.org/user/sd2e/' +
       'UCSB_GBW_LandingSiteDesignv0_1through36/' +
       'UCSB_GBW_LandingSiteDesignv0_1through36_collection/1')

# Fetch this many at a time
step = 5
query = sbol2.SearchQuery(sbol2.SBOL_SEQUENCE, limit=step)
query[sbol2.SBOL_COLLECTION] = uri

# Find out how many there are
count = sbh.search_count_advanced(query)
print(f'There are {count} sequences to pull')

# Now fetch them a chunk at a time
for offset in range(0, count, step):
    print(f'pulling from offset {offset}')
    query.offset = offset
    result = sbh.search_advanced(query)
    identities = [item.identity for item in result]
    sbh.pull(identities, doc, recursive=False)
print(doc)



__version__ = '1.3'

# Anything imported here is part of the public API. Limit what gets
# imported to only those things that are actually needed.
#
# We should be using `__all__` for the list of exported items for this
# module, but the list of items in `constants.py` is extensive. So
# instead of `__all__`, we will try to be careful about what gets
# imported into this file. In the absence of __all__, all imported
# symbols are also exported.
from .attachment import Attachment
from .collection import Collection
from .combinatorialderivation import CombinatorialDerivation, VariableComponent
from .component import Component
from .component import FunctionalComponent
from .componentdefinition import ComponentDefinition
from .config import Config, ConfigOptions
from .config import getHomespace, hasHomespace, setHomespace
from .constants import *
from .dbtl import Design
from .document import Document, IGEM_STANDARD_ASSEMBLY
from .experiment import Experiment, ExperimentalData
from .identified import Identified
from .implementation import Implementation
from .interaction import Interaction
from .location import Location, Range, Cut, GenericLocation
from .measurement import Measurement
from .module import Module
from .moduledefinition import ModuleDefinition
from .object import SBOLObject
from .participation import Participation
from .searchquery import SearchQuery
from .partshop import PartShop
from .property import DateTimeProperty, FloatProperty, IntProperty
from .property import LiteralProperty, OwnedObject
from .property import ReferencedObject, TextProperty
from .property import URIProperty
from .provo import Activity
from .provo import Agent
from .provo import Association
from .provo import Plan
from .provo import Usage
from .sbolerror import SBOLError, SBOLErrorCode
from .sequence import Sequence
from .sequenceannotation import SequenceAnnotation
from .toplevel import TopLevel
from .validation import is_alphanumeric_or_underscore
from .validation import is_not_alphanumeric_or_underscore
from .versionproperty import VersionProperty



from .constants import *
from .identified import Identified
from .location import OwnedLocation
from .property import OwnedObject, URIProperty, ReferencedObject


class SequenceAnnotation(Identified):

    def __init__(self, uri='example', version=VERSION_STRING):
        """Construct a SequenceAnnotation
        :param uri: A full URI including a scheme, namespace, and identifier.
        If SBOLCompliance configuration is enabled, then this argument
        is simply the displayId for the new object and a full URI
        will automatically be constructed.
        :param version: An arbitrary version string.
        If SBOLCompliance is enabled, this should be a Maven version string
        of the form "major.minor.patch".
        """
        super().__init__(SBOL_SEQUENCE_ANNOTATION, uri, version)
        self.component = ReferencedObject(self, SBOL_COMPONENT_PROPERTY,
                                          SBOL_COMPONENT, '0', '1', None)
        self.locations = OwnedLocation(self, SBOL_LOCATIONS,
                                       '0', '*', [])
        self.roles = URIProperty(self, SBOL_ROLES, '0', '*', [])

    def addRole(self, new_role):
        val = self.roles
        val.append(new_role)
        self.roles = val

    def removeRole(self, index=0):
        val = self.roles
        del val[index]
        self.roles = val

    def precedes(self, comparand):
        """Tests if the comparand SequenceAnnotation precedes this one
        according to base coordinates

        :param comparand: Another SequenceAnnotation
        :return: true or false
        """
        raise NotImplementedError("Not yet implemented")

    def follows(self, comparand):
        """Tests if the comparand SequenceAnnotation follows this one
        according to base coordinates

        :param comparand: Another SequenceAnnotation
        :return: true or false
        """
        raise NotImplementedError("Not yet implemented")

    def contains(self, comparand):
        """Tests if the comparand SequenceAnnotation is contained within
        the same start and end base coordinates as this one.
        This is mutually exclusive with overlaps.

        :param comparand: Another SequenceAnnotation
        :return: true or false
        """
        raise NotImplementedError("Not yet implemented")

    def overlaps(self, comparand):
        """Tests if the comparand SequenceAnnotation overlaps with this one
        in the primary sequence

        :param comparand: Another SequenceAnnotation
        :return: true or false
        """
        raise NotImplementedError("Not yet implemented")

    def extract(self, start_reference=1):
        """Convert a SequenceAnnotation to a subcomponent

        :param start_reference:
        :return: A ComponentDefinition representing the subcomponent
        """
        raise NotImplementedError("Not yet implemented")

    def length(self):
        """

        :return: The length of a SequenceAnnotation in base coordinates.
        """
        raise NotImplementedError("Not yet implemented")



import os
import tempfile
import unittest

import sbol2

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
CRISPR_EXAMPLE = os.path.join(MODULE_LOCATION, 'resources', 'crispr_example.xml')


class TestSequence(unittest.TestCase):

    def testAddSequence(self):
        test_seq = sbol2.Sequence("R0010", "ggctgca")
        doc = sbol2.Document()
        doc.addSequence(test_seq)
        seq = doc.sequences.get("R0010").elements

        self.assertEqual(seq, 'ggctgca')

    def testRemoveSequence(self):
        test_seq = sbol2.Sequence("R0010", "ggctgca")
        doc = sbol2.Document()
        doc.addSequence(test_seq)
        doc.sequences.remove(0)
        with self.assertRaises(sbol2.SBOLError):
            doc.sequences.get("R0010")

    def testSeqDisplayId(self):
        doc = sbol2.Document()
        doc.read(CRISPR_EXAMPLE)

        # List of displayIds
        listseq = ['CRP_b_seq', 'CRa_U6_seq', 'gRNA_b_seq', 'mKate_seq']

        listseq_read = [seq.displayId for seq in doc.sequences]

        self.assertCountEqual(listseq_read, listseq)

    def testSequenceEncoding(self):
        doc = sbol2.Document()
        doc.read(CRISPR_EXAMPLE)
        seq = doc.sequences.get('CRP_b_seq')
        self.assertEqual(seq.encoding, sbol2.SBOL_ENCODING_IUPAC)

    def testSequenceElement(self):
        sbol2.setHomespace('http://sbols.org/CRISPR_Example')
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)
        doc = sbol2.Document()
        doc.read(CRISPR_EXAMPLE)
        # Sequence to test against
        seq = ('GCTCCGAATTTCTCGACAGATCTCATGTGATTACGCCAAGCTACGGGCGGAGTACTGTCCTC'
               'CGAGCGGAGTACTGTCCTCCGAGCGGAGTACTGTCCTCCGAGCGGAGTACTGTCCTCCGAGC'
               'GGAGTTCTGTCCTCCGAGCGGAGACTCTAGATACCTCATCAGGAACATGTTGGAATTCTAGG'
               'CGTGTACGGTGGGAGGCCTATATAAGCAGAGCTCGTTTAGTGAACCGTCAGATCGCCTCGAG'
               'TACCTCATCAGGAACATGTTGGATCCAATTCGACC')

        seq_read = doc.sequences.get('CRP_b_seq').elements
        self.assertEqual(seq_read, seq)

    def testUpdateSequenceElement(self):
        sbol2.setHomespace('http://sbols.org/CRISPR_Example')
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)
        doc = sbol2.Document()
        doc.read(CRISPR_EXAMPLE)
        # Sequence to test against
        seq = 'AAAAA'
        doc.sequences.get('CRP_b_seq').elements = seq
        seq_read = doc.sequences.get('CRP_b_seq').elements
        self.assertEqual(seq_read, seq)

    # File I/O Tests
    def testUpdateWrite(self):
        sbol2.setHomespace('http://sbols.org/CRISPR_Example')
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)
        doc = sbol2.Document()
        doc.read(CRISPR_EXAMPLE)
        # Sequence to test against
        seq = 'AAAAA'
        doc.sequences.get('CRP_b_seq').elements = seq
        doc2 = sbol2.Document()  # Document to compare for equality
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, 'test.xml')
            # Write to disk
            doc.write(test_file)
            # Compare
            doc2.read(test_file)
        seq_read = doc2.sequences.get('CRP_b_seq').elements
        self.assertEqual(seq_read, seq)

    def test_bool(self):
        seq = sbol2.Sequence()
        self.assertTrue(seq)



from .constants import *
from .component import FunctionalComponent
from .identified import Identified
from .measurement import Measurement
from .participation import Participation
from .property import OwnedObject
from .property import URIProperty
from . import validation


class Interaction(Identified):
    def __init__(self, uri='example', interaction_type=SBO_INTERACTION,
                 *, type_uri=SBOL_INTERACTION, version=VERSION_STRING):
        super().__init__(type_uri=type_uri, uri=uri, version=version)
        self.functionalComponents = OwnedObject(self,
                                                SBOL_FUNCTIONAL_COMPONENTS,
                                                FunctionalComponent,
                                                '0', '*', [validation.libsbol_rule_18])
        self.types = URIProperty(self, SBOL_TYPES,
                                 '1', '*', [], interaction_type)
        self.participations = OwnedObject(self, SBOL_PARTICIPATIONS,
                                          Participation,
                                          '0', '*', [])
        self.measurements = OwnedObject(self, SBOL_MEASUREMENTS,
                                        Measurement, '0', '*', [])



import sbol2

doc = sbol2.Document()
GFP = sbol2.ComponentDefinition("GFP")
doc.add(GFP)
collection = doc.collections.create("collection")
collection.members = [GFP.identity]
doc.write("Test_Collections.xml")



import collections
from typing import Any


class URIDict(collections.UserDict):

    def __init__(self):
        super().__init__()

    def __getitem__(self, key: str) -> Any:
        return super().__getitem__(str(key))

    def __setitem__(self, key: str, value: Any) -> None:
        super().__setitem__(str(key), value)

    def __delitem__(self, key: str) -> None:
        super().__delitem__(str(key))

    def __contains__(self, key: str) -> bool:
        return super().__contains__(str(key))



import time
import sbol2

# This example is from the documentation at
# https://pysbol.readthedocs.io/en/latest/repositories.html

igem = sbol2.PartShop('https://synbiohub.org/public/igem')
records = []
search_term = 'plasmid'
limit = 25
total_hits = igem.searchCount(search_term)
print(f'Expecting {total_hits} records')
for offset in range(0, total_hits, limit):
    records.extend(igem.search(search_term, sbol2.SBOL_COMPONENT_DEFINITION,
                               offset, limit))
    time.sleep(0.1)
print(f'Received {len(records)} records')



import unittest

import rdflib

import sbol2


class TestCollection(unittest.TestCase):

    def __init__(self, *args):
        super().__init__(*args)
        self.homespace = 'http://example.org'

    def make_identity(self, homespace, name, version, tipe='Collection'):
        uri = '{}/{}/{}/{}'
        return uri.format(homespace, tipe, name, version)

    def setUp(self):
        # Always reset homespace
        sbol2.setHomespace(self.homespace)
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, True)
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_COMPLIANT_URIS, True)

    def testEmptyConstructor(self):
        # This is the default name in the collection constructor
        name = 'example'
        c = sbol2.Collection()
        expected_identity = self.make_identity(self.homespace, name,
                                               sbol2.VERSION_STRING)
        self.assertEqual(c.identity, expected_identity)
        self.assertEqual(c.displayId, name)
        self.assertEqual(c.version, sbol2.VERSION_STRING)
        self.assertEqual(c.rdf_type, sbol2.SBOL_COLLECTION)
        doc = sbol2.Document()
        doc.addCollection(c)
        self.assertEqual(len(doc.collections), 1)

    def test1ArgConstructor(self):
        name = 'foo'
        c = sbol2.Collection(name)
        expected_identity = self.make_identity(self.homespace, name,
                                               sbol2.VERSION_STRING)
        self.assertEqual(c.identity, expected_identity)
        self.assertEqual(c.displayId, name)
        self.assertEqual(c.version, sbol2.VERSION_STRING)
        self.assertEqual(c.rdf_type, sbol2.SBOL_COLLECTION)
        doc = sbol2.Document()
        doc.addCollection(c)
        self.assertEqual(len(doc.collections), 1)

    def test2ArgConstructor(self):
        name = 'foo'
        version = '3'
        c = sbol2.Collection(name, version)
        expected_identity = self.make_identity(self.homespace, name,
                                               version)
        self.assertEqual(c.identity, expected_identity)
        self.assertEqual(c.displayId, name)
        self.assertEqual(c.version, version)
        self.assertEqual(c.rdf_type, sbol2.SBOL_COLLECTION)
        doc = sbol2.Document()
        doc.addCollection(c)
        self.assertEqual(len(doc.collections), 1)

    def test3ArgConstructor(self):
        name = 'foo'
        version = '3'
        rdf_type = sbol2.SBOL_COLLECTION + '2'
        c = sbol2.Collection(name, version, type_uri=rdf_type)
        expected_identity = self.make_identity(self.homespace, name,
                                               version, 'Collection2')
        self.assertEqual(c.identity, expected_identity)
        self.assertEqual(c.displayId, name)
        self.assertEqual(c.version, version)
        self.assertEqual(c.rdf_type, rdf_type)
        # Verify that when added to a document, this collection is not
        # in the list of collections. That's because the rdf_type is
        # not SBOL_COLLECTION.
        doc = sbol2.Document()
        doc.addCollection(c)
        self.assertEqual(len(doc.collections), 0)



#!/usr/bin/env python3

# Filter FASTA files based on genbank information

import argparse
import logging
from pathlib import Path
import sys

from Bio import SeqIO
from Bio.SeqFeature import FeatureLocation, CompoundLocation
from Bio.SeqRecord import SeqRecord


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("fastafile", type=argparse.FileType('r'),
                        metavar="FASTA_FILE")
    parser.add_argument("gbdir", metavar="GENBANK_DIR")
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-o', '--outfile', type=argparse.FileType('a'),
                        default=sys.stdout, metavar="OUTPUT_FILE")
    parser.add_argument('-p', '--padding', type=int, default=0,
                        help="extra padding around genbank regions")
    args = parser.parse_args(args)
    return args


def init_logging(debug=False):
    msgFormat = '%(asctime)s %(levelname)s %(message)s'
    dateFormat = '%m/%d/%Y %H:%M:%S'
    level = logging.INFO
    if debug:
        level = logging.DEBUG
    logging.basicConfig(format=msgFormat, datefmt=dateFormat, level=level)


def fragment_accession(accession_version, fragment):
    accession, version = accession_version.split('.')
    return '{}-fragment{}.{}'.format(accession, fragment, version)


def write_sequence(out_fp, sequence, desc, accession):
    header_fmt = 'gi|NONE|gb|{}|Generated by gbfilter'
    header = header_fmt.format(accession)
    record = SeqRecord(sequence, id=header, description=desc)
    SeqIO.write(record, out_fp, 'fasta')


def feature_description(feature, sub_features=None):
    type_string = f'{feature.type}'
    if sub_features:
        subs = '-'.join(sub_features)
        type_string = f'{type_string} {subs}'
    return f'{type_string} {feature.location}'


def pad_compound_location(location, seq_len, padding):
    logging.debug('CompoundLocation: %r', location)
    if not (len(location.parts) == 2 and location.start == 0 and
            location.end == seq_len):
        logging.warning('compound location does not wrap from end to start: %r',
                        location)
        # raise AssertionError("Compound location does not wrap from end to start")
    new_parts = [pad_location(loc, seq_len, padding) for loc in location.parts]
    new_location = CompoundLocation(new_parts, location.operator)
    return new_location


def pad_location(location, seq_len, padding):
    """Given a FeatureLocation, pad it and return the new location.
    """
    # logging.debug('seq_len = %r', seq_len)
    # logging.debug('padding = %r', padding)
    # logging.debug('location.start = %r', location.start)
    # logging.debug('location.end = %r', location.end)
    new_start = max(0, location.start - padding)
    new_end = min(seq_len, location.end + padding)
    # logging.debug('new_start = %r', new_start)
    # logging.debug('new_end = %r', new_end)
    return FeatureLocation(new_start, new_end, strand=location.strand)


def padded_location(location, seq_len, padding):
    if isinstance(location, CompoundLocation):
        return pad_compound_location(location, seq_len, padding)
    return pad_location(location, seq_len, padding)


def get_filtered_sequences(gb_file, padding):
    # This is a generator, yielding sequences instead of returning a
    # list of all sequences. This way we don't keep a long list in
    # memory when we're just writing them out to the contrast fasta
    # file
    filtered_features = ['tRNA', 'rRNA', 'tmRNA']
    filtered_cds_genes = frozenset(['rpoA', 'RpoA',
                                    'rpoB', 'RpoB',
                                    'rpoC', 'RpoC',
                                    'rpoD', 'RpoD',
                                    'dnaA', 'DnaA',
                                    'dnaB', 'DnaB',
                                    'gyrB', 'GyrB'])
    for record in SeqIO.parse(str(gb_file), 'genbank'):
        for feature in record.features:
            if feature.type in filtered_features:
                feature.location = padded_location(feature.location,
                                                   len(record.seq),
                                                   padding)
                # logging.debug('location.start = %r', feature.location.start)
                # logging.debug('location.end = %r', feature.location.end)
                contrast_seq = feature.extract(record.seq)
                yield contrast_seq, feature_description(feature)
            elif feature.type == 'CDS':
                # Coding Sequence
                gene = set(feature.qualifiers.get('gene', []))
                gene_names = gene.intersection(filtered_cds_genes)
                if not gene_names:
                    continue
                # The gene is one we want to filter
                # (code reuse is copying code so it exists in two places, right?
                feature.location = padded_location(feature.location,
                                                   len(record.seq),
                                                   padding)
                contrast_seq = feature.extract(record.seq)
                yield contrast_seq, feature_description(feature, gene_names)


def append_sequences(gb_file, fasta_fp, accession_version, padding):
    # Given a genbank file (gb_file) and an open fasta file for
    # output, write sequences that should be given to contrast based
    # on rules encoded in get_filtered_sequences. The
    # accession_version of the original sequence is used to generate
    # an accession version for each fragment output to the fasta file.
    i = 0
    for seq, desc in get_filtered_sequences(gb_file, padding):
        i += 1
        accession = fragment_accession(accession_version, i)
        write_sequence(fasta_fp, seq, desc, accession)


def main(argv=None):
    args = parse_args(argv)
    init_logging(args.debug)

    # Convert to a pathlib.Path for easier appending
    gb_path = Path(args.gbdir)

    # read the fasta file
    # for each accession:
    #    if there is a matching gb file
    #        parse the gb file
    #
    for rec in SeqIO.parse(args.fastafile, 'fasta'):
        dfields = rec.description.split('|')
        accession = None
        if dfields[0] == 'gi' and dfields[2] == 'gb':
            accession = dfields[3]
        # Parse other FASTA header types here
        gb_file = gb_path / (accession + '.gb')
        if gb_file.exists():
            logging.debug(gb_file)
            append_sequences(gb_file, args.outfile, accession, args.padding)


if __name__ == '__main__':
    main()



import unittest

import rdflib

import sbol2


class TestAttachment(unittest.TestCase):

    def test_exported(self):
        # Ensure Attachment is exported from sbol2
        self.assertIn('Attachment', dir(sbol2))

    def test_type_uri(self):
        # Test that after construction the attachment has the correct URI.
        # This was a problem until the arguments to the constructor were
        # rearranged.
        test_attach = sbol2.Attachment("TEST")
        self.assertEqual(sbol2.SBOL_ATTACHMENT, test_attach.rdf_type)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import sbol2


MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))

# For testing reading of measurements
MEASURE_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL2',
                                'Measure.xml')


class TestMeasurement(unittest.TestCase):

    def test_add_remove_type(self):
        measurement = sbol2.Measurement()
        self.assertEqual([], measurement.types)
        measurement.addType(sbol2.BIOPAX_DNA)
        self.assertEqual([sbol2.BIOPAX_DNA], measurement.types)
        measurement.addType(sbol2.BIOPAX_RNA)
        self.assertEqual([sbol2.BIOPAX_DNA, sbol2.BIOPAX_RNA], measurement.types)
        measurement.addType(sbol2.BIOPAX_COMPLEX)
        expected = [sbol2.BIOPAX_DNA, sbol2.BIOPAX_RNA, sbol2.BIOPAX_COMPLEX]
        self.assertEqual(expected, measurement.types)
        measurement.removeType(1)
        self.assertEqual([sbol2.BIOPAX_DNA, sbol2.BIOPAX_COMPLEX], measurement.types)

    def test_read(self):
        doc = sbol2.Document(MEASURE_LOCATION)
        md_uri = 'http://www.async.ece.utah.edu/md'
        md = doc.getModuleDefinition(md_uri)
        interaction_uri = 'http://www.async.ece.utah.edu/md/i'
        interaction = md.interactions.get(interaction_uri)
        measurement_uri = 'http://www.async.ece.utah.edu/md/i/i_measure'
        measurement = interaction.measurements.get(measurement_uri)
        self.assertIsNotNone(measurement)
        # Expecting this to be a Measurement, for starters
        # This wasn't the case, see issue #304
        self.assertTrue(isinstance(measurement, sbol2.Measurement))
        self.assertEqual(0.04, measurement.value)
        self.assertEqual('http://purl.obolibrary.org/obo/UO_0000077', measurement.unit)


if __name__ == '__main__':
    unittest.main()



import logging

import sbol2

# This example shows how to configure logging for sbol2


# Configure the Python logging system
logging.basicConfig()

# Set the 'sbol2' logger to debug level
logging.getLogger('sbol2').setLevel(logging.DEBUG)

# This will generate lots of debug logging
doc = sbol2.Document('data/BBa_I0462.xml')



from rdflib import URIRef

from .toplevel import TopLevel
from .constants import *
from .property import URIProperty


class Collection(TopLevel):

    def __init__(self, uri=URIRef("example"), version=VERSION_STRING,
                 *, type_uri=SBOL_COLLECTION):
        super().__init__(type_uri, uri, version)
        self.members = URIProperty(self, SBOL_MEMBERS, '0', '*', [])



import sbol2

doc = sbol2.Document()
igem = sbol2.PartShop('https://synbiohub.org/public/igem')
igem.pull('BBa_I719005', doc)

composition = doc.componentDefinitions.create('composition')
composition.assemblePrimaryStructure(['BBa_I719005', 'BBa_I719005'])
doc.write("minimal_example.xml")



import time
import timeit
import rdflib

EXAMPLE_PATH = '../test/SBOLTestSuite/SBOL2/RepressionModel.xml'
EXAMPLE_PATH = '../W303.xml'


def load_document(path):
    graph = rdflib.Graph()
    graph.parse(location=path, format="application/rdf+xml")
    return graph


start = time.time()
seconds = timeit.timeit(lambda: load_document(EXAMPLE_PATH), number=1)
end = time.time()
print('timeit result: {}'.format(seconds))
print('elapsed: {}'.format(end-start))



from .constants import *
from .identified import Identified
from .property import URIProperty, FloatProperty

from rdflib import URIRef


class Measurement(Identified):
    """The purpose of the Measure class is to link a numerical value
    to a unit of measure."""

    def __init__(self, uri=URIRef('example'), value=0.0,
                 unit='', version=VERSION_STRING):
        super().__init__(OM_MEASURE, uri, version)
        self.value = FloatProperty(self, OM_HAS_NUMERICAL_VALUE,
                                   '1', '1', [], value)
        self.unit = URIProperty(self, OM_HAS_UNIT, '1', '1', [], unit)
        self.types = URIProperty(self, SBOL_TYPES, '0', '*', [])

    def addType(self, new_type):
        val = self.types
        val.append(new_type)
        self.types = val

    def removeType(self, index=0):
        val = self.types
        del val[index]
        self.types = val



import logging
import posixpath
from typing import List

from deprecated import deprecated
import rdflib
from rdflib import URIRef

from .config import getHomespace, string_equal
from .config import hasHomespace
from .config import Config
from .constants import *
from .property import Property, OwnedObject, URIProperty
from .sbolerror import SBOLError
from .sbolerror import SBOLErrorCode
from .uridict import URIDict
from . import SBOL2Serialize
from . import validation


# This is an internal function, not part of the public API
def _compare_properties(obj1, obj2):
    obj1_keys = obj1.properties.keys()
    if sorted(obj1_keys) != sorted(obj2.properties.keys()):
        return False
    # Keys are equal, check values by converting to sets
    for k in obj1_keys:
        if set(obj1.properties[k]) != set(obj2.properties[k]):
            return False
    return True


# This is an internal function, not part of the public API
def _compare_owned_objects(obj1, obj2):
    obj1_keys = obj1.owned_objects.keys()
    if sorted(obj1_keys) != sorted(obj2.owned_objects.keys()):
        return False
    # Keys are equal, check values by converting to dicts
    for k in obj1_keys:
        # Recursively compare child objects
        oo1 = {oo.identity: oo for oo in obj1.owned_objects[k]}
        oo2 = {oo.identity: oo for oo in obj2.owned_objects[k]}
        if oo1.keys() != oo2.keys():
            return False
        for uri in oo1:
            if not oo1[uri].compare(oo2[uri]):
                return False
    return True


class SBOLObject:
    """An SBOLObject converts a Python data structure into an RDF triple store
     and contains methods for serializing and parsing RDF triples.
    """

    def _serialize(self):
        # Convert and SBOL object into RDF triples.
        raise NotImplementedError("Not yet implemented")

    def _nest(self):
        # Pretty-writer that converts flat RDF/XML into nested RDF/XML
        # (ie. SBOL)
        raise NotImplementedError("Not yet implemented")

    def _makeQName(self, uri):
        raise NotImplementedError("Not yet implemented")

    # TODO Docstrings on variables isn't a thing in Python. Consider using Epydoc.
    # The identity property is REQUIRED by all Identified objects
    # and has a data type of URI. A given Identified object's identity
    # URI MUST be globally unique among all other identity URIs.
    # The identity of a compliant SBOL object MUST begin with
    # a URI prefix that maps to a domain over which the user has control.
    # Namely, the user can guarantee uniqueness of identities
    # within this domain. For other best practices regarding URIs
    # see Section 11.2 of the [SBOL specification document
    # (http://sbolstandard.org/wp-content/uploads/2015/08/SBOLv2.0.1.pdf).

    def __init__(self, type_uri=rdflib.URIRef(UNDEFINED),
                 uri=rdflib.URIRef("example")):
        """Open-world constructor."""
        self.owned_objects = URIDict()  # map<rdf_type, vector<SBOLObject>>
        self.properties = URIDict()  # map<rdf_type, vector<SBOLObject>>
        self.doc = None
        self.parent = None
        self._default_namespace = None
        self._hidden_properties = []
        self.rdf_type = str(type_uri)
        self._namespaces = {}
        self.identity = URIProperty(self, SBOL_IDENTITY, '0', '1',
                                    [validation.sbol_rule_10202])
        uri = URIRef(uri)
        if hasHomespace():
            uri = posixpath.join(getHomespace(), uri)
        self.identity = uri

    @property
    def logger(self):
        logger = logging.getLogger('sbol2')
        if not logger.hasHandlers():
            # If there are no handlers, nobody has initialized
            # logging.  Configure logging here so we have a chance of
            # seeing the messages.
            logging.basicConfig()
        return logger

    def __uri__(self) -> rdflib.URIRef:
        """Returns a URIRef representing this object.

        :return: The identity of this object as an rdflib.URIRef
        :rtype: rdflib.URIRef
        """
        return rdflib.URIRef(self.identity)

    @property
    def type(self):
        return self.rdf_type

    @property
    @deprecated(reason='`this` no longer makes sense')
    def this(self):
        # For backward compatibility with pySBOL. Just return myself.
        return self

    def getTypeURI(self):
        """
        :return: The uniform resource identifier that describes
        the RDF-type of this SBOL Object.
        """
        return self.rdf_type

    def getClassName(self, rdf_type):
        """Parses a local class name from the RDF-type of this SBOL Object."""
        if '#' in rdf_type:
            return rdf_type.split('#')[1]
        else:
            return rdf_type

    @deprecated(reason='This is an obsolete carryover from the C++ library')
    def cast(self, cls):
        if not isinstance(self, cls):
            msg = 'Cannot cast instance of {} to class {}'
            msg = msg.format(type(self).__name__, cls.__name__)
            raise TypeError(msg)
        # TODO: libSBOL does a copy here. I don't think that's needed
        # in Python.
        return self

    def find(self, uri):
        """Search this object recursively to see if an object or
        any child object with URI already exists.

        :param uri: The URI to search for.
        :return: The SBOLObject associated with this URI if it exists,
        None otherwise.
        """
        uri = rdflib.URIRef(uri)
        if string_equal(self.identity, uri):
            return self
        for rdf_type, object_store in self.owned_objects.items():
            if rdf_type in self._hidden_properties:
                continue
            for obj in object_store:
                match = obj.find(uri)
                # NOTE: original implementation has 2 recursive calls,
                # which is probably bad...
                if match is not None:
                    return match
        return None

    def cacheObjects(self, objectCache):
        """TODO document

        :param objectCache: a dictionary mapping strings to SBOLObjects
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def find_property(self, uri):
        """Search this object recursively to see if it contains
        a member property with the given RDF type.

        :param uri: The RDF type of the property to search for.
        :return: The SBOLObject that contains a member property
        with the specified RDF type, None otherwise
        """
        if uri in self.owned_objects and uri in self.properties:
            return self
        for rdf_type, object_store in self.owned_objects.items():
            for obj in object_store:
                match = obj.find_property(uri)
                if match:
                    return match
        return None

    def find_property_value(self, uri, value, matches=None):
        """Search this object recursively to see if it contains
        a member property with the given RDF type
        and indicated property value.

        :param uri: The RDF type of the property to search for.
        :param value: The property value to match.
        :param matches:
        :return: A vector containing all objects found that contain
        a member property with the specified RDF type
        """
        matches = []
        for rdf_type, store in self.owned_objects.items():
            # Skip hidden and aliased properties
            if rdf_type in self._hidden_properties:
                continue
            for obj in store:
                matches += obj.find_property_value(uri, value)
        try:
            value_store = self.properties[uri]
            for val in value_store:
                if string_equal(val, value):
                    matches.append(self)
        except KeyError:
            # It is ok that uri is not in properties
            pass
        return matches

    def find_reference(self, uri):
        """Search this object recursively to see if it contains
        a member property with the given RDF type
        and indicated property value.

        :param uri: A URI, either an ontology term or an object reference,
        to search for.
        :return: A vector containing all objects found that contain
        the URI in a property value.
        """
        references = []
        # Ask all the owned objects
        for oo_list in self.owned_objects.values():
            for oo in oo_list:
                references.extend(oo.find_reference(uri))
        # Check myself
        for pvals in self.properties.values():
            if uri in pvals:
                references.append(self)
        return references

    def compare(self, other):
        """Compare two SBOLObjects. The behavior is currently undefined for
        objects with custom annotations or extension classes.

        :param other: The object being compared to this one.
        :return: True if the objects are identical, False if they are different.

        """
        if type(other) != type(self):
            return False
        if self.rdf_type != other.rdf_type:
            return False
        if not _compare_properties(self, other):
            return False
        if not _compare_owned_objects(self, other):
            return False
        return True

    def getPropertyValue(self, property_uri: str) -> str:
        """Get the value of a custom annotation property by its URI.

        :param property_uri: The URI for the property.
        :type property_uri: str
        :return: The value of the property.
        :rtype: str
        :raises: SBOLError(SBOL_ERROR_NOT_FOUND) if property does not exist
        """
        values = self.getPropertyValues(property_uri)
        try:
            return str(values[0])
        except IndexError:
            return ''

    def getPropertyValues(self, property_uri: str) -> List[str]:
        """Get all values of a custom annotation property by its URI.

        :param property_uri: The URI for the property.
        :type property_uri: str
        :return: A vector of property values.
        :rtype: list of strings
        :raises: SBOLError(SBOL_ERROR_NOT_FOUND) if property does not exist
        """
        try:
            return [str(x) for x in self.properties[property_uri]]
        except KeyError as e:
            # no property by this name
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            'Property {} not found'.format(property_uri))

    def getProperties(self):
        """Gets URIs for all properties contained by this object.
        This includes SBOL core properties as well as
        custom annotations. Use this to find custom extension data
        in an SBOL file.

        :return: A vector of URIs that identify the properties contained in this object.
        """
        property_uris = [p for p in self.properties.keys()]
        owned_objects_uris = [p for p in self.owned_objects.keys()]
        return property_uris + owned_objects_uris

    @deprecated(reason='Use extension properties instead')
    def setPropertyValue(self, property_uri, val):
        """Set and overwrite the value for a user-defined annotation property.

        :param property_uri:
        :param val: A string
        :return: None
        """
        if not isinstance(val, str):
            raise TypeError('%r is not a string', val)
        # Ensure that the property is a URIRef
        property_uri = rdflib.URIRef(property_uri)
        # If there is effectively no value (i.e. '') clear out the
        # value
        if not val:
            # No value, no property, do nothing
            if property_uri not in self.properties:
                return
            # If a value exists in the property, set only the first
            # entry to the empty string. This is backward compatible
            # with pySBOL/libSBOL
            if self.properties[property_uri]:
                self.properties[property_uri][0] = rdflib.Literal('')
            return
        # If the value does not exist, just set it and we're done
        if property_uri not in self.properties:
            self.properties[property_uri] = [rdflib.Literal(val)]
            return
        # Ensure that val is a Literal or URIRef
        if not isinstance(val, (rdflib.Literal, rdflib.URIRef)):
            # Default to Literal
            val = rdflib.Literal(val)
            # If an existing property is a URIRef, convert val to URIRef
            if self.properties[property_uri]:
                if isinstance(self.properties[property_uri][0], rdflib.URIRef):
                    val = rdflib.URIRef(val)
        self.properties[property_uri][0] = val

    def addPropertyValue(self, property_uri, val):
        """Append a value to a user-defined annotation property.

        :param property_uri:
        :param val: Either a literal or URI value.
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def setAnnotation(self, property_uri, val):
        """Set the value for a user-defined annotation property.
        Synonymous with setPropertyValue.

        :param property_uri:
        :param val: If the value is a URI,
        it should be surrounded by angle brackets,
        else it will be interpreted as a literal value.
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def getAnnotation(self, property_uri):
        """Get the value of a custom annotation property by its URI.
        Synonymous with getPropertyValue.

        :param property_uri: The URI for the property.
        :return: The value of the property or SBOL_ERROR_NOT_FOUND.
        """
        raise NotImplementedError("Not yet implemented")

    def apply(self, callback, user_data):
        """
        Recursively call this callback function on the user_data provided
        :param callback: the callback function
        :param user_data: the input to the callback
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def update_uri(self):
        """
        Recursively generates SBOL compliant ids for an object and all
        its owned objects, then checks to make sure that these ids are unique.
        :return: None
        """
        raise NotImplementedError("Implemented by child classes")

    def build_graph(self, graph):
        graph.add((rdflib.URIRef(self.identity),
                   rdflib.RDF.type,
                   rdflib.URIRef(self.rdf_type)))
        for typeURI, proplist in self.properties.items():
            if typeURI in self._hidden_properties:
                # Skip hidden properties
                continue
            if typeURI == SBOL_IDENTITY:
                # Do not write the identity property to SBOL files
                continue
            for prop in proplist:
                graph.add((rdflib.URIRef(self.identity),
                           rdflib.URIRef(typeURI),
                           prop))
        for typeURI, objlist in self.owned_objects.items():
            if typeURI in self._hidden_properties:
                # Skip hidden properties
                continue

            for owned_obj in objlist:
                graph.add((rdflib.URIRef(self.identity),
                           rdflib.URIRef(typeURI),
                           URIRef(owned_obj.identity)))
                owned_obj.build_graph(graph)

            # register ownership relationship in SBOL2Serialize to structure XML
            SBOL2Serialize.register_ownership_relation(self.getTypeURI(),
                                                       typeURI)

    def __str__(self):
        return self.identity

    def is_top_level(self):
        return False

    def __getattribute__(self, name):
        # Call the default method
        result = object.__getattribute__(self, name)
        if isinstance(result, OwnedObject):
            if result.getUpperBound() == '1':
                if result:
                    result = result[0]
                else:
                    result = None
        elif isinstance(result, Property):
            # Else if attribute is any other kind of Property besides
            # OwnedObject, convert so the Property attributes are
            # transparent and look like native types.
            result = result.value
        return result

    def _is_transparent_attribute(self, name):
        try:
            attr = self.__dict__[name]
        except KeyError:
            return False
        return isinstance(attr, Property)

    def _set_transparent_attribute(self, name, value):
        self.__dict__[name].set(value)

    def __setattr__(self, name, value):
        if self._is_transparent_attribute(name):
            self._set_transparent_attribute(name, value)
            return
        object.__setattr__(self, name, value)

    def _added_to_document(self, doc):
        """Classes can override this method to do extra work when
        being added to a document."""
        pass



import os
import posixpath
import unittest

import sbol2

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL_TEST_SUITE = os.path.join(MODULE_LOCATION, 'SBOLTestSuite')


class MyTestCase(unittest.TestCase):

    def test_exported(self):
        self.assertIn('CombinatorialDerivation', dir(sbol2))
        self.assertIn('VariableComponent', dir(sbol2))

    def test_read(self):
        fname = os.path.join(SBOL_TEST_SUITE,
                             'SBOL2_ic/gfp_reporter_template.xml')
        doc = sbol2.Document(fname)
        uri = 'http://michael.zhang/GFP_Reporter_Template_CombinatorialDerivation'
        cd: sbol2.CombinatorialDerivation = doc.find(uri)
        self.assertIsNotNone(cd)
        self.assertEqual(1, len(cd.variableComponents))
        vc_uri = posixpath.join(uri, 'CDS_Component_VariableComponent')
        vc: sbol2.VariableComponent = doc.find(vc_uri)
        self.assertEqual(vc.identity, cd.variableComponents[0].identity)
        self.assertIsInstance(vc, sbol2.VariableComponent)
        self.assertEqual(3, len(vc.variants))

    def test_read2(self):
        fname = os.path.join(SBOL_TEST_SUITE,
                             'SBOL2_ic/eukaryotic_transcriptional_unit.xml')
        doc = sbol2.Document(fname)
        uri = posixpath.join('http://michael.zhang',
                             'Eukaryotic_Transcriptional_Unit_CombinatorialDerivation')
        cd: sbol2.CombinatorialDerivation = doc.find(uri)
        self.assertIsNotNone(cd)
        self.assertEqual(2, len(cd.variableComponents))
        vc_uri = posixpath.join(uri, 'Pro_Component_VariableComponent')
        vc: sbol2.VariableComponent = doc.find(vc_uri)
        self.assertIsNotNone(vc)
        self.assertIsInstance(vc, sbol2.VariableComponent)
        self.assertEqual(1, len(vc.variantDerivations))


if __name__ == '__main__':
    unittest.main()



from .constants import *
from .identified import Identified
from .mapsto import MapsTo
from .measurement import Measurement
from .property import OwnedObject, ReferencedObject


class Module(Identified):
    def __init__(self, uri='example',
                 # Everything after the asterisk (`*`) is a keyword
                 *, definition='', version=VERSION_STRING,
                 type_uri=SBOL_MODULE):
        super().__init__(type_uri, uri, version)
        self.definition = ReferencedObject(self, SBOL_DEFINITION,
                                           SBOL_MODULE_DEFINITION,
                                           '1', '1', [], definition)
        self.mapsTos = OwnedObject(self, SBOL_MAPS_TOS, MapsTo,
                                   '0', '*', [])
        self.measurements = OwnedObject(self, SBOL_MEASUREMENTS,
                                        Measurement, '0', '*', [])



import datetime
from abc import ABC, abstractmethod
import collections
import logging
import math
import posixpath
from typing import Any, Union

import dateutil.parser
from rdflib import Literal, URIRef
import packaging.version as pv

from .config import Config
from .config import ConfigOptions
from .config import getHomespace
from .config import parseClassName
from .config import parsePropertyName
from .config import string_equal
from .constants import *
from .sbolerror import SBOLError
from .sbolerror import SBOLErrorCode


def sort_version(obj):
    return obj.version


class Property(ABC):
    """Member properties of all SBOL objects are defined
    using a Property object.

    The Property class provides a generic interface for accessing SBOL objects.
    At a low level, the Property class converts SBOL data structures
    into RDF triples.
    """

    @staticmethod
    def valid_lower_bound(x: Union[int, float, str]) -> str:
        """Validate the lower bound. Allow numeric strings, ints,
        and floats.
        """
        if isinstance(x, str) and x.isnumeric():
            return x
        elif isinstance(x, int):
            return str(x)
        elif isinstance(x, float) and x.is_integer():
            return str(int(x))
        else:
            raise ValueError('Lower bound must be numeric')

    @staticmethod
    def valid_upper_bound(x: Union[int, float, str]) -> str:
        """Validate the upper bound. Allow numeric strings, ints,
        floats, and '*' (to represent infinity).
        """
        if x == '*':
            return x
        if x is math.inf:
            return '*'
        try:
            x = Property.valid_lower_bound(x)
        except ValueError:
            raise ValueError('Upper bound must be numeric or \'*\'')
        return x

    def __init__(self, property_owner, type_uri, lower_bound, upper_bound,
                 validation_rules, initial_value=None):
        """Construct a Property.

        :param property_owner: All Property objects must have a pointer back
        to their parent SBOLObject.
        :param type_uri: An RDF hash URI for this property, consisting
        of a namespace followed by an identifier.
        For example, Properties of SBOLObjects use URIs of the form
        http://sbols.org/v2#somePropertyName, where the identifier
        somePropertyName determines the appearance of XML nodes
        in an SBOL file. Alternatively, annotations in a custom namespace
        can be provided using a similarly formed hash URI
        in another namespace.
        :param lower_bound:
        :param upper_bound:
        :param validation_rules: A vector of externally defined ValidationRules
        The vector contains pointers to functions which correspond to
        the validation rules listed in the appendix of the formal
        SBOL specification document.  ValidationRules are automatically
        checked every time a setter or adder method is called and
        when Documents are read and written.
        :param initial_value: The initial value of the Property
        (int, str, float supported)
        """
        # Ensure property_owner has a properties attribute of type
        # dict. This is a "fail early" check.
        if not isinstance(property_owner.properties, collections.abc.Mapping):
            raise TypeError('property_owner.properties must be a dict')
        self._sbol_owner = property_owner
        self._rdf_type = str(type_uri)
        self._lowerBound = Property.valid_lower_bound(lower_bound)
        self._upperBound = Property.valid_upper_bound(upper_bound)
        # Validate validation rules
        if validation_rules is None:
            # Some constructors pass None for validation rules
            # Convert to empty list
            validation_rules = []
        for vr in validation_rules:
            if not callable(vr):
                raise TypeError('Validation rule %r is not callable' % vr)
        self._validation_rules = validation_rules
        if initial_value is not None:
            self.value = initial_value

    @property
    def logger(self):
        logger = logging.getLogger('sbol2')
        if not logger.hasHandlers():
            # If there are no handlers, nobody has initialized
            # logging.  Configure logging here so we have a chance of
            # seeing the messages.
            logging.basicConfig()
        return logger

    def getTypeURI(self):
        """

        :return: URI representing the predicate.
        """
        return self._rdf_type

    def getOwner(self):
        """

        :return: The owner of this Property.
        """
        return self._sbol_owner

    def getRawValue(self):
        if self._rdf_type not in self._sbol_owner.properties:
            return None
        properties = self._sbol_owner.properties[self._rdf_type]
        if self._upperBound == '1':
            try:
                return properties[-1]
            except IndexError:
                # List is empty
                return None
        # upperbound is not 1, just return the list
        return properties

    @property
    def value(self):
        raise NotImplementedError("getter is only implemented by subclasses")

    @value.setter
    def value(self, new_value):
        self.set(new_value)

    @abstractmethod
    def set(self, new_value):
        raise NotImplementedError("set() is only implemented by subclasses")

    @abstractmethod
    def add(self, new_value):
        """Appends the new value to a list of values,
        for properties that allow it."""
        raise NotImplementedError("add() is only implemented by subclasses")

    def remove(self, index):
        """Remove a property value. By default, we assume this is a literal
        located at index 0.
        """
        if self._sbol_owner is not None:
            if self._rdf_type in self._sbol_owner.properties:
                properties = self._sbol_owner.properties[self._rdf_type]
                if index >= len(properties):
                    raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT,
                                    'Index out of range')
                if len(properties) == 1:
                    self.clear()
                else:
                    del properties[index]
        else:
            self.logger.error("Unable to update property: SBOL owner not set.")

    def clear(self):
        """Clear all property values."""
        properties = self._sbol_owner.properties[self._rdf_type]
        properties.clear()

    def write(self):
        """Write property values."""
        subject = self._sbol_owner.identity.get()
        predicate = self._rdf_type
        obj = self.value
        print('Subject: ' + subject)
        print('Predicate: ' + predicate)
        print('Object: ' + obj)

    def find(self, query):
        """Check if a value in this property matches the query."""
        msg = '{}.find() not yet implemented for query {} of type {}'
        msg = msg.format(type(self).__name__, query, type(query))
        raise NotImplementedError(msg)

    def getLowerBound(self):
        return self._lowerBound

    @property
    def lower_bound(self):
        """Return the lower bound of the property as an integer.

        """
        return int(self._lowerBound)

    def getUpperBound(self):
        return self._upperBound

    @property
    def upper_bound(self):
        """Return the upper bound of the property as a number.  If the upper
        bound is '*', this value is math.inf. Otherwise the upper
        bound is converted to an integer and returned.

        """
        if self._upperBound == '*':
            return math.inf
        return int(self._upperBound)

    def validate(self, arg):
        for validation_rule in self._validation_rules:
            validation_rule(self._sbol_owner, arg)

    def __contains__(self, item):
        try:
            obj = self.find(item)
        except SBOLError as err:
            if err.error_code() == SBOLErrorCode.NOT_FOUND_ERROR:
                return None
            if err.error_code() == SBOLErrorCode.SBOL_ERROR_NOT_FOUND:
                return None
        else:
            return bool(obj)

    def _isHidden(self):
        return self._rdf_type in self._sbol_owner._hidden_properties

    def __len__(self):
        if self._rdf_type not in self._sbol_owner.properties:
            return 0
        else:
            properties = self._sbol_owner.properties[self._rdf_type]
            return len(properties)

    def __str__(self):
        if self._rdf_type not in self._sbol_owner.properties:
            return ""
        else:
            return str(self._sbol_owner.properties[self._rdf_type])

    @staticmethod
    def guess_args(validators, initial_value):
        """In pysbol there are 4 constructors for each Property. We have
        to figure out which constructor the user intended to call based on
        the arguments. The signatures are:
            IntProperty(owner, rdf_type, low_bound, high_bound, ValidationRules,
                        initial_value)
            IntProperty(owner, rdf_type, low_bound, high_bound, ValidationRules)
            IntProperty(owner, rdf_type, low_bound, high_bound, initial_value)
            IntProperty(owner, rdf_type, low_bound, high_bound)

        We have to guess which of the two 5-arg constructors the user
        intended to call. If both validators and initial_value are None,
        this is the 4-arg constructor, so return None for both. If neither
        validators nor initial_value are None, this is the 6-arg constructor
        so return the values as specified.

        For the 5-arg cases, if the isinstance(initial_value, typeinfo) then
        guess that hte 5th arg is an initial_value. Otherwise assume the 5th
        arg is a list of validation rules.
        """
        if validators is None and initial_value is None:
            # 4-arg constructor
            return None, None
        if initial_value is not None:
            # 6-arg constructor
            return validators, initial_value
        if not isinstance(validators, collections.abc.Iterable):
            # validation rules must be an iterable, like a list,
            # so guess that it is an initial value
            return None, validators
        if all(callable(x) for x in validators):
            # All the values are callable, so guess this is a list of validators
            return validators, None
        # If the 5th arg is not validators, guess that it is an initial value
        return None, validators


class URIProperty(Property):

    def __init__(self, property_owner, type_uri, lower_bound, upper_bound,
                 validation_rules, initial_value=None):
        super().__init__(property_owner, type_uri, lower_bound, upper_bound,
                         validation_rules)
        if self._rdf_type not in self._sbol_owner.properties:
            self._sbol_owner.properties[self._rdf_type] = []
        if initial_value is not None:
            self.validate(initial_value)
            self.value = initial_value

    @property
    def value(self):
        if self._upperBound == '1':
            return self.getSinglePropertyValue()
        else:
            return self.getPropertyValueList()

    def getSinglePropertyValue(self):
        if self._rdf_type not in self._sbol_owner.properties:
            return None
        properties = self._sbol_owner.properties[self._rdf_type]
        if len(properties) == 0:
            return None
        return self.convert_to_user(properties[-1])

    def getPropertyValueList(self):
        if self._rdf_type not in self._sbol_owner.properties:
            return []
        store = self._sbol_owner.properties[self._rdf_type]
        return [self.convert_to_user(x) for x in store]

    @value.setter
    def value(self, new_value):
        self.set(new_value)

    def set(self, new_value):
        self.validate(new_value)
        if self.getUpperBound() == '1':
            self.setSinglePropertyValue(new_value)
        else:
            self.setPropertyValueList(new_value)

    def setSinglePropertyValue(self, new_value):
        new_value = self.convert_from_user(new_value)
        # clear out any old value
        self._sbol_owner.properties[self._rdf_type].clear()
        if new_value is None:
            # We've already cleared the value, do nothing else.
            return
        self._sbol_owner.properties[self._rdf_type].append(new_value)

    def setPropertyValueList(self, value):
        if value is None:
            self._sbol_owner.properties[self._rdf_type] = []
            return
        if isinstance(value, str):
            value = [value]
        if isinstance(value, collections.abc.Iterable):
            new_value = [self.convert_from_user(v) for v in value]
            self._sbol_owner.properties[self._rdf_type] = new_value
            return
        # Unrecognized value. Raise type error
        msg = 'Cannot set URIProperty to value {} of type {}'
        raise TypeError(msg.format(value, type(value).__name__))

    def add(self, new_value):
        # Note: this is only called internally. This method is not
        # accessible from outside the library via the public API.
        self.value += [new_value]

    def convert_to_user(self, value):
        result = str(value)
        if result == '':
            # special case, empty strings are equivalent to None
            return None
        return result

    def convert_from_user(self, value):
        # None is ok iff upper bound is 1 and lower bound is 0.
        # If upper bound > 1, attribute is a list and None is not a valid list
        # If lower bound > 0, attribute must have a value, so None is unacceptable
        if value is None and self.upper_bound == 1 and self.lower_bound == 0:
            return None
        if not isinstance(value, str):
            msg = '{} values must have type str'.format(self.getTypeURI())
            raise TypeError(msg)
        return URIRef(value)


class LiteralProperty(Property):

    def __init__(self, property_owner, type_uri, lower_bound, upper_bound,
                 validation_rules=None, initial_value=None):
        validation_rules, initial_value = Property.guess_args(validation_rules,
                                                              initial_value)
        super().__init__(property_owner, type_uri, lower_bound, upper_bound,
                         validation_rules)
        if self._rdf_type not in self._sbol_owner.properties:
            self._sbol_owner.properties[self._rdf_type] = []
        if initial_value is not None:
            self.value = initial_value

    @property
    def value(self):
        if self._upperBound == '1':
            return self.getSinglePropertyValue()
        else:
            return self.getPropertyValueList()

    def getSinglePropertyValue(self):
        properties = self._sbol_owner.properties[self._rdf_type]
        if len(properties) == 0:
            return None
        # Just return the object by itself (not a list)
        return self.convert_to_user(properties[0])

    def getPropertyValueList(self):
        properties = self._sbol_owner.properties[self._rdf_type]
        return [self.convert_to_user(v) for v in properties]

    @value.setter
    def value(self, new_value):
        self.set(new_value)

    def set(self, new_value):
        if self.getUpperBound() == '1':
            self.setSinglePropertyValue(new_value)
        else:
            self.setPropertyValueList(new_value)

    def setSinglePropertyValue(self, new_value):
        new_value = self.convert_from_user(new_value)
        # clear out any old value
        self._sbol_owner.properties[self._rdf_type].clear()
        if new_value is None:
            # We've already cleared the value, do nothing else.
            return
        self._sbol_owner.properties[self._rdf_type].append(new_value)

    def setPropertyValueList(self, new_value):
        if not isinstance(new_value, collections.abc.Iterable):
            raise TypeError('{} must be an iterable'.format(self.getTypeURI()))
        # Special case. Should we really support this?
        # Convert a string to a list of that string
        if isinstance(new_value, str):
            new_value = [new_value]
        new_value = [self.convert_from_user(v) for v in new_value]
        # should we filter out None in the list? I don't think so. If
        # the user put them there, then that's what they're going to
        # get.
        self._sbol_owner.properties[self._rdf_type] = new_value

    def add(self, new_value):
        property = self._sbol_owner.properties[self._rdf_type]
        if len(property) >= self.upper_bound:
            msg = 'Property {} has reached the upper bound of {}'
            msg = msg.format(self._rdf_type, self.upper_bound)
            raise ValueError(msg)
        new_value = self.convert_from_user(new_value)
        property.append(new_value)

    def convert_to_user(self, value):
        return str(value)

    def convert_from_user(self, value):
        # None is ok iff upper bound is 1 and lower bound is 0.
        # If upper bound > 1, attribute is a list and None is not a valid list
        # If lower bound > 0, attribute must have a value, so None is unacceptable
        if value is None and self.upper_bound == 1 and self.lower_bound == 0:
            return None
        if not isinstance(value, str):
            msg = '{} values must have type str'.format(self.getTypeURI())
            raise TypeError(msg)
        return Literal(value)


class IntProperty(LiteralProperty):

    def convert_to_user(self, value):
        return int(value)

    def convert_from_user(self, value):
        # None is ok iff upper bound is 1 and lower bound is 0.
        # If upper bound > 1, attribute is a list and None is not a valid list
        # If lower bound > 0, attribute must have a value, so None is unacceptable
        if value is None and self.upper_bound == 1 and self.lower_bound == 0:
            return None
        # Convert to int. If conversion fails a ValueError is raised
        value = int(value)
        return Literal(value)


class FloatProperty(LiteralProperty):

    def convert_to_user(self, value: rdflib.Literal) -> float:
        return float(value)

    def convert_from_user(self, value: Any) -> Union[rdflib.Literal, None]:
        """
        :raises: ValueError if value cannot be converted to float
        """
        # None is ok iff upper bound is 1 and lower bound is 0.
        # If upper bound > 1, attribute is a list and None is not a valid list
        # If lower bound > 0, attribute must have a value, so None is unacceptable
        if value is None and self.upper_bound == 1 and self.lower_bound == 0:
            return None
        # Convert to float. If conversion fails a ValueError is raised
        value = float(value)
        return Literal(value)


class DateTimeProperty(LiteralProperty):

    def convert_to_user(self, value):
        return dateutil.parser.parse(value)

    def convert_from_user(self, value):
        # None is ok iff upper bound is 1 and lower bound is 0.
        # If upper bound > 1, attribute is a list and None is not a valid list
        # If lower bound > 0, attribute must have a value, so None is unacceptable
        if value is None and self.upper_bound == 1 and self.lower_bound == 0:
            return None
        if isinstance(value, str):
            value = dateutil.parser.parse(value)
        if not isinstance(value, datetime.datetime):
            msg = '{} values must have type datetime'.format(self.getTypeURI())
            raise ValueError(msg)
        return Literal(value)


class TextProperty(LiteralProperty):

    # In the future, pull the convert_to_user and convert_from_user
    # methods out of LiteralProperty and into TextProperty. Then make
    # LiteralProperty an abstract base class.

    def convert_to_user(self, value):
        result = str(value)
        if result == '':
            # special case, empty strings are equivalent to None
            return None
        return result


class OwnedObject(Property):
    def __init__(self, property_owner, sbol_uri, builder, lower_bound, upper_bound,
                 validation_rules=None, first_object=None):
        """Initialize a container and optionally put the first object in it.
        If validation rules are specified, they will be checked upon initialization.

        builder is a function that takes a single argument, a string,
        and constructs an object of appropriate type for this
        OwnedObject instance. For instance, if this OwnedObject is
        intended to hold ComponentDefinitions, then the builder should
        return an object of type ComponentDefinition.

        """
        super().__init__(property_owner, sbol_uri, lower_bound, upper_bound,
                         validation_rules, first_object)
        self.validate(first_object)
        if not callable(builder):
            msg = '{!r} object is not callable'
            raise TypeError(msg.format(type(builder)))
        self.builder = builder
        # Register Property in owner Object
        if self._sbol_owner is not None:
            self._sbol_owner.properties.pop(self._rdf_type, None)
            self._sbol_owner.owned_objects[self._rdf_type] = []  # vector of SBOLObjects
            if first_object is not None:
                self._sbol_owner.owned_objects[self._rdf_type].append(first_object)

    def create(self, uri, builder=None):
        """Creates an instance appropriate for this owned object collection.

        uri - the name of the object

        """
        if not builder:
            builder = self.builder
        else:
            # Override the default builder to create a subtype of the builder, e.g.,
            # create a Range instead of a Location
            if not callable(builder):
                msg = '{!r} object is not callable'
                raise TypeError(msg.format(type(builder)))
            if not issubclass(builder, self.builder):
                msg = '{!r} is not a subclass of {!r}'
                raise TypeError(msg.format(type(builder), type(self.builder)))
        builder_args = dict(uri=uri)
        if self._sbol_owner and hasattr(self._sbol_owner, 'version'):
            builder_args['version'] = self._sbol_owner.version
        obj = builder(**builder_args)
        self.add(obj)
        return obj

    def add(self, sbol_obj):
        if self._sbol_owner is None:
            # Just silently do nothing?
            return

        # If this is a top level object, add it and all its children recursively to the
        # Document. (With some additional refactoring, this could probably all be handled
        # from this method, thus eliminating need for a separate Document adder method)
        if sbol_obj.is_top_level() and self._sbol_owner.doc is not None:
            self._sbol_owner.doc.add(sbol_obj)
            # If a property is hidden, don't return yet, because it still needs to be
            # added as a child object. (By definition, a hidden owned object can be
            # accessed from both the Document top level and as a child of another top
            # level)
            if not self._isHidden():
                return

        # Not top level, add to the attribute
        object_store = self._sbol_owner.owned_objects[self._rdf_type]
        if sbol_obj in object_store:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE,
                            "The object " + sbol_obj.identity +
                            " is already contained by the " +
                            self._rdf_type + " property")
        # Add to Document and check for uniqueness of URI
        if self._sbol_owner.doc is not None:
            sbol_obj.doc = self._sbol_owner.doc
        sbol_obj.parent = self._sbol_owner
        # Update URI for the argument object and all its children,
        # if SBOL-compliance is enabled.
        sbol_obj.update_uri()
        # Check that this URI is unique within the object store
        # See issue #127
        for obj in object_store:
            if obj.identity == sbol_obj.identity:
                raise SBOLError(SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE,
                                "The object " + sbol_obj.identity +
                                " is already contained by the " +
                                self._rdf_type + " property")
        # Add to parent object
        object_store.append(sbol_obj)
        # Run validation rules
        self.validate(sbol_obj)

    def __getitem__(self, id):
        if type(id) is int:
            return self.get_int(id)
        if isinstance(id, str):
            return self.get_uri(rdflib.URIRef(id))
        # Anything other than int and str, covered above, is a problem
        errmsg = 'id must be str or int, got {!r}'.format(type(id))
        raise TypeError(errmsg)

    def get_int(self, id):
        object_store = self._sbol_owner.owned_objects[self._rdf_type]
        if id >= len(object_store):
            # Note: for loops expect an IndexError to be
            # raised to properly detect end of sequence
            raise IndexError
        return object_store[id]

    def find(self, query):
        if isinstance(query, str):
            try:
                return self.get_uri(rdflib.URIRef(query))
            except SBOLError as e:
                if e.error_code() == SBOLErrorCode.NOT_FOUND_ERROR:
                    return False
                raise
        errmsg = 'Invalid find query {} of type {}'
        errmsg = errmsg.format(query, type(query).__name__)
        raise TypeError(errmsg)

    def get_uri(self, id):
        id = str(id)
        if Config.getOption(ConfigOptions.VERBOSE.value) is True:
            print('SBOL compliant URIs are set to ' +
                  Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value))
            print('SBOL typed URIs are set to ' +
                  Config.getOption(ConfigOptions.SBOL_TYPED_URIS.value))
            print('Searching for ' + id)
        # Search this property's object store for the uri
        object_store = self._sbol_owner.owned_objects[self._rdf_type]
        for obj in object_store:
            if id == obj.identity:
                return obj
        # Now assume the search string is a persistent identity
        obj = self.find_persistent_identity(id)
        if obj is not None:
            return obj
        # If searching by the full URI fails, assume the user is searching
        # for an SBOL-compliant URI using the displayId only
        # Form compliant URI for child object
        parent_obj = self._sbol_owner
        resource_namespaces = []
        resource_namespaces.append(getHomespace())
        if parent_obj.doc is not None:
            for ns in parent_obj.doc.resource_namespaces:
                resource_namespaces.append(ns)
        # Check for regular, SBOL-compliant URIs
        obj = self.find_resource(id, resource_namespaces,
                                 object_store, parent_obj, typedURI=False)
        if obj is not None:
            return obj
        else:
            obj = self.find_resource(id, resource_namespaces,
                                     object_store, parent_obj, typedURI=True)
            if obj is not None:
                return obj
            else:
                msg = 'Object {} not found'.format(id)
                raise SBOLError(SBOLErrorCode.NOT_FOUND_ERROR, msg)

    def find_persistent_identity(self, search_uri):
        if not Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS):
            # Must be using compliant URIs to search by persistent identity
            return None
        # Search for persistent identity, returning the newest version
        # Similar to the looping in find_resource, but we can do it better
        # if a match, store it. If another match, check versions and keep
        # the newer one.
        found_object = None
        found_version = pv.NegativeInfinity
        object_store = self._sbol_owner.owned_objects[self._rdf_type]
        for obj in object_store:
            if obj.identity.startswith(search_uri):
                obj_version = pv.parse(obj.version)
                if obj_version > found_version:
                    found_object = obj
                    found_version = obj_version
        return found_object

    def find_resource(self, uri, resource_namespaces, object_store,
                      parent_obj, typedURI=False):
        persistentIdentity = ''
        for ns in resource_namespaces:
            # Assume the parent object is TopLevel and form the compliant URI
            #
            # Note the final empty string, which causes a trailing '/'
            # to be added. Not sure why we need the trailing slash,
            # but that's why the '' is there.
            if typedURI is True:
                compliant_uri = posixpath.join(ns, parseClassName(self._rdf_type),
                                               uri, '')
            else:
                compliant_uri = posixpath.join(ns, uri, '')
            compliant_uri = URIRef(compliant_uri)
            persistent_id_matches = []
            if Config.getOption(ConfigOptions.VERBOSE) is True:
                print('Searching for TopLevel: ' + compliant_uri)
            for obj in object_store:
                if compliant_uri in obj.identity:
                    persistent_id_matches.append(obj)
                # Sort objects with same persistentIdentity by version
                # TODO is this right?
                persistent_id_matches.sort(key=sort_version)
            # If objects matching the persistentIdentity were found,
            # return the most recent version
            if len(persistent_id_matches) > 0:
                return persistent_id_matches[-1]
            # Assume the object is not TopLevel # TODO What is this for?
            if (SBOL_PERSISTENT_IDENTITY in parent_obj.properties
                    and parent_obj.properties[SBOL_PERSISTENT_IDENTITY]):
                persistentIdentity = parent_obj.properties[SBOL_PERSISTENT_IDENTITY][0]
            if (SBOL_VERSION in parent_obj.properties
                    and parent_obj.properties[SBOL_VERSION]):
                version = parent_obj.properties[SBOL_VERSION][0]
                compliant_uri = posixpath.join(persistentIdentity, uri, version)
            else:
                compliant_uri = posixpath.join(persistentIdentity, uri)
            if Config.getOption(ConfigOptions.VERBOSE.value) is True:
                print('Searching for non-TopLevel: ' + compliant_uri)
            for obj in object_store:
                if string_equal(obj.identity, compliant_uri):
                    return obj

    def get(self, uri=''):
        # TODO: orig getter contains a size check when uri is a constant string
        if uri == '':
            object_store = self._sbol_owner.owned_objects[self._rdf_type]
            if object_store:
                return object_store[0]
            else:
                return None
        else:
            return self.__getitem__(uri)

    def __setitem__(self, rdf_type, sbol_obj):
        # NOTE: custom implementation. Not sure where this is defined
        # in the original code.
        if self._sbol_owner.is_top_level():
            doc = self._sbol_owner.doc
            if self._isHidden() and doc.find(sbol_obj.identity):
                # In order to avoid a duplicate URI error, don't attempt
                # to add the object if this is a hidden property,
                pass
            else:
                doc.add(sbol_obj)
        # Add to parent object
        if len(self._sbol_owner.owned_objects[rdf_type]) == 0:
            self._sbol_owner.owned_objects[rdf_type].append(sbol_obj)
        else:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT,
                            "Cannot set " + parsePropertyName(rdf_type) +
                            " property. The property is already set. "
                            "Call remove before attempting to "
                            "overwrite the value.")
        sbol_obj.parent = self._sbol_owner
        # Update URI for the argument object and all its children,
        # if SBOL-compliance is enabled.
        sbol_obj.update_uri()

        # Run validation rules
        self.validate(sbol_obj)

    def convert_to_user(self, value):
        return str(value)

    def getSinglePropertyValue(self):
        if self._rdf_type not in self._sbol_owner.properties:
            return None
        properties = self._sbol_owner.properties[self._rdf_type]
        if len(properties) == 0:
            return None
        return self.convert_to_user(properties[-1])

    def getPropertyValueList(self):
        if self._rdf_type not in self._sbol_owner.properties:
            return []
        store = self._sbol_owner.properties[self._rdf_type]
        return [self.convert_to_user(x) for x in store]

    @property
    def value(self):
        if self._upperBound == '1':
            return self.getSinglePropertyValue()
        else:
            return self.getPropertyValueList()

    @value.setter
    def value(self, new_value):
        if new_value is not None:
            self.set(new_value)

    def set(self, new_value):
        if self._sbol_owner is None:
            # Not sure we should raise this, but I'm guessing it would
            # be good to know.
            # TODO: convert this to a better Python exception or to an
            # SBOL exception
            raise Exception('No owner for referenced value')
        if self.getUpperBound() == '1':
            self.setSinglePropertyValue(new_value)
        else:
            self.setPropertyValueList(new_value)

    def setSinglePropertyValue(self, new_value):
        # Clear out the previous value, then add the new value
        #
        # TODO: This can leave the attribute empty if `add` fails.
        # Can we capture that and sent the old value back again?
        if new_value is None:
            value = self.get()
            if value is not None:
                self.remove(value.identity)
            return
        self._sbol_owner.owned_objects[self._rdf_type].clear()
        self.add(new_value)

    def setPropertyValueList(self, new_value):
        # Clear out the previous value, then add the new value
        #
        # TODO: This can leave the attribute empty if `add` fails.
        # Can we capture that and sent the old value back again?
        self._sbol_owner.owned_objects[self._rdf_type].clear()
        for nv in new_value:
            self.add(nv)

    def remove(self, identifier):
        """id can be either an integer index or a string URI"""
        if type(identifier) is int:
            return self.removeOwnedObject_int(identifier)
        elif isinstance(identifier, str):
            return self.removeOwnedObject_str(identifier)
        else:
            msg = 'id parameter must be an integer index or a string uri.'
            msg += ' Got {} of type {}'.format(identifier,
                                               type(identifier).__name__)
            raise TypeError(msg)

    def removeOwnedObject_int(self, index):
        if self._sbol_owner is not None:
            if self._rdf_type in self._sbol_owner.owned_objects:
                object_store = self._sbol_owner.owned_objects[self._rdf_type]
                if index >= len(object_store):
                    raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT,
                                    "Index out of range")
                obj = object_store[index]
                if self._sbol_owner.getTypeURI() == SBOL_DOCUMENT:
                    del obj.doc.SBOLObjects[rdflib.URIRef(obj.identity)]
                del object_store[index]
                obj.doc = None
                self.validate(None)
                return obj
        else:
            raise Exception('This property is not defined in '
                            'the parent object')

    def removeOwnedObject_str(self, uri):
        if not self._sbol_owner:
            return
        obj = self.find(uri)
        object_store = self._sbol_owner.owned_objects[self._rdf_type]
        object_store.remove(obj)
        # Erase TopLevel objects from Document
        if self._sbol_owner.rdf_type == SBOL_DOCUMENT:
            del obj.doc.SBOLObjects[obj.identity]
        obj.doc = None
        self.validate(None)
        return obj

    def clear(self):
        if self._sbol_owner is not None:
            if self._rdf_type in self._sbol_owner.owned_objects:
                object_store = self._sbol_owner.owned_objects[self._rdf_type]
                for obj in object_store:
                    if obj.is_top_level() and obj.doc is not None:
                        obj.doc.SBOLObjects.remove(obj.identity)
                object_store.clear()

    def __len__(self):
        if self._rdf_type not in self._sbol_owner.owned_objects:
            return 0
        else:
            object_store = self._sbol_owner.owned_objects[self._rdf_type]
            return len(object_store)


class ReferencedObject(URIProperty):
    def __init__(self, property_owner, type_uri, reference_type_uri,
                 lower_bound, upper_bound, validation_rules,
                 initial_value=None):
        super().__init__(property_owner, type_uri, lower_bound, upper_bound,
                         validation_rules, initial_value)
        self.reference_type_uri = reference_type_uri
        if self._sbol_owner is not None:
            property_store = []
            self._sbol_owner.properties[type_uri] = property_store

    def get(self, item):
        """Return reference object at this index.
        :param item (int) the index
        """
        reference_store = self._sbol_owner.properties[self._rdf_type]
        return reference_store[item]

    def set(self, new_value):
        if self._sbol_owner is None:
            # Not sure we should raise this, but I'm guessing it would
            # be good to know.
            # TODO: convert this to a better Python exception or to an
            # SBOL exception
            raise Exception('No owner for referenced value')
        if self.getUpperBound() == '1':
            self.setSinglePropertyValue(new_value)
        else:
            self.setPropertyValueList(new_value)

    def _to_uri(self, obj):
        """Converts strings, URIRefs, and sbol.Identified instances into
        rdflib.URIRefs.

        """
        if isinstance(obj, str):
            return rdflib.URIRef(obj)
        if hasattr(obj, '__uri__'):
            # SBOLObjects have a __uri__ method, and others can too
            return obj.__uri__()
        errmsg = 'Cannot convert {} of type {} to a referenced object'
        raise TypeError(errmsg.format(obj, type(obj)))

    def setSinglePropertyValue(self, new_value):
        if new_value is None or new_value == '':
            self._sbol_owner.properties[self._rdf_type] = []
            return
        self._sbol_owner.properties[self._rdf_type] = [self._to_uri(new_value)]

    def setPropertyValueList(self, new_value):
        if new_value is None:
            self._sbol_owner.properties[self._rdf_type] = []
            return
        if hasattr(new_value, '__uri__'):
            # Convert to URI
            # SBOLObjects have a __uri__ method, and others can too
            new_value = new_value.__uri__()
        if isinstance(new_value, str):
            # Turn it into a list
            new_value = [new_value]
        if isinstance(new_value, collections.abc.Iterable):
            # Convert the items to URIRefs
            new_value = list([self._to_uri(x) for x in new_value])
        self._sbol_owner.properties[self._rdf_type] = new_value

    def add(self, uri):
        # Does anyone call this method? If so, it needs to be properly
        # implemented.
        raise Exception('Should ReferencedObject.add be deprecated?')
        if self._sbol_owner is not None:
            self._sbol_owner.properties[self._rdf_type].append(uri)
        else:
            # NOTE: we could raise an exception here, but
            # the original code is not doing anything in this case.
            print('Unable to set item. SBOL owner was None.')

    def addReference(self, uri):
        # Does anyone call this method? If so, it needs to be properly
        # implemented.
        raise Exception('Should ReferencedObject.addReference be deprecated?')
        self._sbol_owner.properties[self._rdf_type].append(uri)

    def _ensure_uri_ref(self, thing):
        """This method ensures that what gets returned is an rdflib.URIRef. If
        the set methods are working properly, this should be
        unnecessary. Hence the warning message if what it finds is not
        a URIRef. The goal is to remove this method eventually.

        """
        if isinstance(thing, rdflib.URIRef):
            return thing
        self.logger.warning('ReferencedObject was not a URIRef: {}'.format(thing))
        return rdflib.URIRef(thing)

    @property
    def value(self):
        raw_value = self.getRawValue()
        if raw_value is None:
            return None
        if self._upperBound == '1':
            return self.convert_to_user(raw_value)
        else:
            # Return a list, not a generator. Consumers are expecting a list.
            return [self.convert_to_user(rval) for rval in raw_value]

    @value.setter
    def value(self, new_value):
        self.set(new_value)



import unittest

import sbol2
from sbol2.validation import sbol_rule_10202


class TestValidation(unittest.TestCase):

    def test_is_alphanumeric_or_underscore(self):
        # is alpha
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('A'))
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('a'))
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('M'))
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('m'))
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('Z'))
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('z'))
        # is numeric
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('0'))
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('3'))
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('9'))
        # is underscore
        self.assertTrue(sbol2.is_alphanumeric_or_underscore('_'))
        # not alpha or numeric or underscore
        self.assertFalse(sbol2.is_alphanumeric_or_underscore('$'))
        self.assertFalse(sbol2.is_alphanumeric_or_underscore('\n'))
        self.assertFalse(sbol2.is_alphanumeric_or_underscore('-'))
        # More to ensure the function is exported than anything else
        self.assertTrue(sbol2.is_not_alphanumeric_or_underscore('-'))
        self.assertFalse(sbol2.is_not_alphanumeric_or_underscore('A'))

    def test_sbol_rule_10202(self):
        with self.assertRaises(TypeError):
            sbol_rule_10202(None, None)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol2


GSOC_SBH_URL = 'https://synbiohub.org'


class TestSearchQuery(unittest.TestCase):

    def test_search_query1(self):
        query = sbol2.SearchQuery()
        query[sbol2.SBOL_TYPES] = sbol2.BIOPAX_DNA
        self.assertIn(sbol2.SBOL_TYPES, query.properties)
        self.assertEqual(sbol2.BIOPAX_DNA, query[sbol2.SBOL_TYPES])

    def test_gsoc_example_1_title(self):
        # See issue #240
        collection = 'https://synbiohub.org/public/igem/igem_collection/1'
        title = 'GFP'
        query = sbol2.SearchQuery()
        query[sbol2.SBOL_COLLECTION] = collection
        query[sbol2.SBOL_NAME] = title
        # GSOC is always looking for DNA Region
        query[sbol2.SBOL_TYPES] = sbol2.BIOPAX_DNA
        part_shop = sbol2.PartShop(GSOC_SBH_URL)
        response = part_shop.search(query)
        # At least one item in return should be the
        # expected return: https://synbiohub.org/public/igem/BBa_E0040/1
        identities = [r.identity for r in response]
        self.assertIn('https://synbiohub.org/public/igem/BBa_E0040/1', identities)
        #
        # All items in response should have name == GFP exactly.
        names = [r.name == title for r in response]
        self.assertTrue(all(names))

    def test_gsoc_example_1_display_id(self):
        # See issue #240
        collection = 'https://synbiohub.org/public/igem/igem_collection/1'
        display_id = 'BBa_E0040'
        query = sbol2.SearchQuery()
        query[sbol2.SBOL_COLLECTION] = collection
        query[sbol2.SBOL_DISPLAY_ID] = display_id
        # GSOC is always looking for DNA Region
        query[sbol2.SBOL_TYPES] = sbol2.BIOPAX_DNA
        part_shop = sbol2.PartShop(GSOC_SBH_URL)
        response = part_shop.search(query)
        # At least one item in return should be the
        # expected return: https://synbiohub.org/public/igem/BBa_E0040/1
        identities = [r.identity for r in response]
        self.assertIn('https://synbiohub.org/public/igem/BBa_E0040/1', identities)
        #
        # All items in response should have name == GFP exactly.
        display_ids = [r.displayId == display_id for r in response]
        self.assertTrue(all(display_ids))

    def gsoc_search_title(self, part_shop, collection, title):
        query = sbol2.SearchQuery()
        query[sbol2.SBOL_COLLECTION] = collection
        query[sbol2.SBOL_NAME] = title
        # GSOC is always looking for DNA Region
        query[sbol2.SBOL_TYPES] = sbol2.BIOPAX_DNA
        return part_shop.search(query)

    def gsoc_search_display_id(self, part_shop, collection, display_id):
        query = sbol2.SearchQuery()
        query[sbol2.SBOL_COLLECTION] = collection
        query[sbol2.SBOL_DISPLAY_ID] = display_id
        # GSOC is always looking for DNA Region
        query[sbol2.SBOL_TYPES] = sbol2.BIOPAX_DNA
        return part_shop.search(query)

    def do_gsoc_search(self, collection, term):
        part_shop = sbol2.PartShop(GSOC_SBH_URL)
        results = self.gsoc_search_title(part_shop, collection, term)
        results.extend(self.gsoc_search_display_id(part_shop, collection, term))
        return results

    def test_gsoc_example_2(self):
        # Looking for: https://synbiohub.org/public/igem/BBa_R0010/1
        #              and/or https://synbiohub.org/public/igem/BBa_C0012/1
        # Collection: https://synbiohub.org/public/igem/igem_collection/1
        # Search term: "LacI"
        # Don't want: https://synbiohub.org/public/igem/BBa_S03334/1
        #             or https://synbiohub.org/public/igem/BBa_K1444015/1
        collection = 'https://synbiohub.org/public/igem/igem_collection/1'
        results = self.do_gsoc_search(collection,
                                      'LacI')
        identities = [r.identity for r in results]
        self.assertIn('https://synbiohub.org/public/igem/BBa_R0010/1', identities)
        # This item requires case-insensitive search. I don't see a way to do that.
        # self.assertIn('https://synbiohub.org/public/igem/BBa_C0012/1', identities)
        self.assertNotIn('https://synbiohub.org/public/igem/BBa_S03334/1', identities)
        self.assertNotIn('https://synbiohub.org/public/igem/BBa_K1444015/1', identities)

    def test_gsoc_example_3(self):
        # Looking for: https://synbiohub.org/public/igem/BBa_R0010/1
        # Collection: https://synbiohub.org/public/igem/igem_collection/1
        # Search term: "BBa_R0010"
        # Don't want: https://synbiohub.org/public/igem/BBa_S03334/1
        #             or https://synbiohub.org/public/igem/BBa_K1444015/1
        collection = 'https://synbiohub.org/public/igem/igem_collection/1'
        results = self.do_gsoc_search(collection,
                                      'BBa_R0010')
        identities = [r.identity for r in results]
        self.assertIn('https://synbiohub.org/public/igem/BBa_R0010/1', identities)
        self.assertNotIn('https://synbiohub.org/public/igem/BBa_S03334/1', identities)
        self.assertNotIn('https://synbiohub.org/public/igem/BBa_K1444015/1', identities)

    def test_gsoc_example_4(self):
        # Looking for: https://synbiohub.org/public/igem/BBa_B0034/1
        # Collection: https://synbiohub.org/public/igem/igem_collection/1
        # Search term: "BBa_B0034"
        # Don't want: https://synbiohub.org/public/igem/BBa_I13617/1
        collection = 'https://synbiohub.org/public/igem/igem_collection/1'
        results = self.do_gsoc_search(collection,
                                      'BBa_B0034')
        identities = [r.identity for r in results]
        self.assertIn('https://synbiohub.org/public/igem/BBa_B0034/1', identities)
        self.assertNotIn('https://synbiohub.org/public/igem/BBa_I13617/1', identities)

    def test_gsoc_example_5(self):
        # Looking for: https://synbiohub.org/public/bsu/BO_28874/1
        # Collection: https://synbiohub.org/public/bsu/bsu_collection/1
        # Search term: "acca"
        # Don't want: https://synbiohub.org/public/bsu/BO_26604/1
        collection = 'https://synbiohub.org/public/bsu/bsu_collection/1'
        results = self.do_gsoc_search(collection,
                                      # If we can do case-insensitive, use 'acca'
                                      'accA')
        identities = [r.identity for r in results]
        self.assertIn('https://synbiohub.org/public/bsu/BO_28874/1', identities)
        self.assertNotIn('https://synbiohub.org/public/bsu/BO_26604/1', identities)

    def test_gsoc_count_5(self):
        part_shop = sbol2.PartShop(GSOC_SBH_URL)
        query = sbol2.SearchQuery()
        collection = 'https://synbiohub.org/public/bsu/bsu_collection/1'
        query[sbol2.SBOL_COLLECTION] = collection
        query[sbol2.SBOL_NAME] = 'accA'
        # GSOC is always looking for DNA Region
        query[sbol2.SBOL_TYPES] = sbol2.BIOPAX_DNA
        count = part_shop.searchCount(query)
        self.assertEqual(1, count)


if __name__ == '__main__':
    unittest.main()



from rdflib import URIRef

from .toplevel import TopLevel
from .property import URIProperty
from .constants import *


class Model(TopLevel):
    def __init__(self, type_uri=SBOL_MODEL, uri=URIRef('example'), source='',
                 language=EDAM_SBML, framework=SBO_CONTINUOUS,
                 version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        self.source = URIProperty(self, SBOL_SOURCE,
                                  '0', '1', [], source)
        self.language = URIProperty(self, SBOL_LANGUAGE,
                                    '0', '1', [], language)
        self.framework = URIProperty(self, SBOL_FRAMEWORK,
                                     '0', '1', [], framework)



import unittest

from sbol2 import *


class TestIntegrate(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        setHomespace('https://example.com')

    def setUp(self):
        self.doc = Document()
        self.wt_cd = ComponentDefinition('wt_cd')
        self.wt_cd.sequence = Sequence('wt_seq')
        self.insert_cd = ComponentDefinition('insert_cd')
        self.insert_cd.sequence = Sequence('insert_seq')
        self.integrated_cd = ComponentDefinition('integrated_cd')
        self.doc.addComponentDefinition([self.integrated_cd, self.insert_cd, self.wt_cd])

    def testSourceLocation(self):
        # Test insertion splits the targt ComponentDefinition
        # into 2 Components each with a SourceLocation
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 3)
        components = [c for c in self.integrated_cd.components
                      if c.definition == self.wt_cd.identity]
        self.assertEqual(len(components), 2)
        self.assertEqual(len(components[0].sourceLocations), 1)
        self.assertEqual(len(components[1].sourceLocations), 1)
        l0 = components[0].sourceLocations.getRange()
        l1 = components[1].sourceLocations.getRange()
        self.assertEqual(l0.start, 1)
        self.assertEqual(l0.end, 2)
        self.assertEqual(l1.start, 3)
        self.assertEqual(l1.end, 4)

    def testIntegrationPrimaryStructure(self):
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 3)
        primary_structure = self.integrated_cd.getPrimaryStructure()
        primary_structure = [c.identity for c in primary_structure]
        self.assertListEqual(primary_structure, [self.wt_cd.identity,
                                                 self.insert_cd.identity,
                                                 self.wt_cd.identity])

    def testIntegrationNegative(self):
        # An exception should be raised if user tries to insert
        # at a negative base coordinate
        with self.assertRaises(ValueError):
            self.wt_cd.sequence.elements = 'atcg'
            self.insert_cd.sequence.elements = 'gg'
            self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, -3)

    def testIntegrationPositive(self):
        # An exception should be raised if the user attempts to
        # appending at a location beyond the end of the initial
        # sequence.
        with self.assertRaises(ValueError):
            self.wt_cd.sequence.elements = 'atcg'
            self.insert_cd.sequence.elements = 'gg'
            self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 8)

    def testIntegration0(self):
        # An exception should be raised if user tries to insert
        # at 0 base coordinate. Base coordinates are indexed from 1
        with self.assertRaises(ValueError):
            self.wt_cd.sequence.elements = 'atcg'
            self.insert_cd.sequence.elements = 'gg'
            self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 0)

    def testIntegration1(self):
        # Test inserting at the beginning of the sequence. Prepending.
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 1)
        self.integrated_cd.compile()
        self.assertEqual(self.integrated_cd.sequence.elements, 'ggatcg')

    def testIntegration2(self):
        # Test inserting within the sequence.
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 2)
        self.integrated_cd.compile()
        self.assertEqual(self.integrated_cd.sequence.elements, 'aggtcg')

    def testIntegration3(self):
        # Test inserting within the sequence.
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 3)
        self.integrated_cd.compile()
        self.assertEqual(self.integrated_cd.sequence.elements, 'atggcg')

    def testIntegration4(self):
        # Test inserting one base before the end of the sequence.
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 4)
        self.integrated_cd.compile()
        self.assertEqual(self.integrated_cd.sequence.elements, 'atcggg')

    def testIntegrationN(self):
        # Test appending at the end of the initial sequence.
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 5)
        self.integrated_cd.compile()
        self.assertEqual(self.integrated_cd.sequence.elements, 'atcggg')

    def testMissingDocument(self):
        # Test that compile fails with a raised exception when
        # this CD or the insert is not associated with a Document
        self.wt_cd.sequence.elements = 'atcg'
        self.insert_cd.sequence.elements = 'gg'
        self.wt_cd = self.doc.componentDefinitions.remove(self.wt_cd.identity)
        with self.assertRaises(ValueError):
            self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 4)
        self.doc.componentDefinitions.add(self.wt_cd)
        self.insert_cd = self.doc.componentDefinitions.remove(self.insert_cd.identity)
        with self.assertRaises(ValueError):
            self.integrated_cd.integrateAtBaseCoordinate(self.wt_cd, self.insert_cd, 4)



import posixpath

from rdflib import URIRef

from .config import Config
from .config import ConfigOptions
from .config import getHomespace
from .constants import *
from .identified import Identified
from .property import ReferencedObject


class TopLevel(Identified):
    """All SBOL classes derived from TopLevel appear as top level nodes
    in the RDF/XML document tree and SBOL files."""

    def __init__(self, type_uri=SBOL_TOP_LEVEL,
                 uri=URIRef("example"), version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        # TODO could this be moved into 'identified' constructor?
        if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
            if Config.getOption(ConfigOptions.SBOL_TYPED_URIS.value) is True:
                self.persistentIdentity = posixpath.join(getHomespace(),
                                                         self.getClassName(type_uri),
                                                         self.displayId)
        self.attachments = ReferencedObject(self, SBOL_ATTACHMENTS,
                                            SBOL_ATTACHMENT, '0', '*', [], [])

    def addToDocument(self, document):
        raise NotImplementedError("Not yet implemented")

    def generateTopLevel(self, uri, agent=None, plan=None, usages=None):
        """
        :param uri: A URI for the new object, or a displayId if operating
                    in SBOLCompliant mode
        :return:
        """
        # TODO this originally was called 'generate' but
        # TODO it didn't override parent function of same name.
        raise NotImplementedError("Not yet implemented")

    def initialize(self, uri):
        raise NotImplementedError("Not yet implemented")

    def update_uri(self):
        """
        Recursively generates SBOL compliant ids for an object and all
        its owned objects, then checks to make sure that these ids are unique.
        :return: None
        """
        return

    def is_top_level(self):
        return True
